#include "config.h"
#include "ggap/app.h"
#include "ggap/dialogs.h"
#include "ggap/wswindow.h"
#include "ggap/proc.h"
#include "ggap/gap.h"
#include "ggap/prefsdialog.h"
#include "ggap/help.h"
#include "ggap/opts.h"
#include "ggap/script.h"
#include <moo/macros.h>
#include <moo/docmanager.h>
#include <QtGui>
#include <string.h>
#ifdef Q_OS_MAC
#include <CoreFoundation/CoreFoundation.h>
#endif
#ifdef Q_WS_X11
#include <X11/Xlib.h>
#include <QX11Info>
#endif

using namespace ggap;

namespace {

struct AppOptions {
    QStringList files_to_open;
    bool open_help_browser;
    QString help_url;
    bool raise;
    bool unit_tests;
    QStringList unit_test_args;
    QStringList scripts;

    AppOptions() :
        open_help_browser(false),
        raise(false),
        unit_tests(false)
    {
    }
};

static AppOptions appOptions;

#define APP_MSG_VERSION "1.0"

struct AppMsg {
    QString version;
    QStringList files;
    QStringList scripts;
    QString help_url;
    bool open_help_browser;
};

static QDataStream &operator << (QDataStream &s, const AppMsg &msg)
{
    return s << msg.version <<
                msg.files <<
                msg.scripts <<
                msg.help_url <<
                msg.open_help_browser;
}

static QDataStream &operator >> (QDataStream &s, AppMsg &msg)
{
    s >> msg.version;

    if (s.status() != QDataStream::Ok)
        return s;

    if (msg.version != APP_MSG_VERSION)
    {
        s.setStatus(QDataStream::ReadCorruptData);
        return s;
    }

    s >> msg.files;
    s >> msg.scripts;
    s >> msg.help_url;
    s >> msg.open_help_browser;
    return s;
}

}

void App::parseOptions(int &argc, char **argv)
{
    OptParser opts("ggap");

    opts.add("fatal-warnings", 0);
    opts.add("fatal-criticals", 0);

#ifdef Q_OS_MAC
    opts.add("raise", "raise window", &appOptions.raise);
#endif

    opts.add("help-url", "url", "open url in help browser", &appOptions.help_url);
    opts.add("help-browser", "open help browser", &appOptions.open_help_browser);

    opts.add(Option("script", "script", "run script", &appOptions.scripts, Option::NeedValue));

    bool show_version = false;
    opts.add("version", "show version and exit", &show_version);

    opts.parse(argc, argv);

    if (show_version)
    {
        std::cout << "ggap " GGAP_VERSION << std::endl;
        ::exit(EXIT_SUCCESS);
    }

    appOptions.files_to_open = opts.files();
}

QByteArray App::serializeCommandLine()
{
    AppMsg data = {APP_MSG_VERSION,
                   QStringList(),
                   appOptions.scripts,
                   appOptions.help_url,
                   appOptions.open_help_browser};

    foreach (const QString &fn, appOptions.files_to_open)
    {
        QUrl url = QUrl(fn);
        QString file = url.toLocalFile();
        if (!file.isEmpty() && QFileInfo(file).isRelative())
            data.files << QDir::cleanPath(QDir::current().filePath(fn));
        else
            data.files << fn;
    }

    QByteArray msg;
    QDataStream stream(&msg, QIODevice::WriteOnly);
    stream << data;
    return QByteArray("appmsg:") + msg.toBase64();
}

void App::doAppMsg(const QByteArray &data)
{
    AppMsg msg;

    QByteArray decoded = QByteArray::fromBase64(data);
    QDataStream stream(&decoded, QIODevice::ReadOnly);
    stream >> msg;

    if (stream.status() != QDataStream::Ok)
    {
        qCritical("%s: got bad data", Q_FUNC_INFO);
        return;
    }

    foreach (const QString &script, msg.scripts)
        runScript(script);

    if (msg.scripts.isEmpty() && msg.files.isEmpty() &&
        msg.help_url.isEmpty() && !msg.open_help_browser)
            forceActivate();

    foreach (const QString &f, msg.files)
    {
        loadFile(f);
        forceActivate();
    }

    if (!msg.help_url.isEmpty() || msg.open_help_browser)
    {
        HelpBrowser::showWindow(msg.help_url);
        forceActivate();
    }
}

static QWidget *topWindow()
{
    if (QWidget *window = QApplication::activeWindow())
        return window;

    foreach (QWidget *w, QApplication::topLevelWidgets())
        if (w->isWindow())
            return w;

    return 0;
}

#ifdef Q_WS_X11
static void activateWindowX11(QWidget *window)
{
    static Atom activeWindowAtom = XInternAtom(QX11Info::display(), "_NET_ACTIVE_WINDOW", false);

    XClientMessageEvent e;
    memset(&e, 0, sizeof e);
    e.type = ClientMessage;
    e.window = window->winId();
    e.message_type = activeWindowAtom;
    e.format = 32;
    e.data.l[0] = 2;
    e.data.l[1] = CurrentTime;
    e.data.l[2] = 0;
    e.data.l[3] = 0;
    e.data.l[4] = 0;

    XSendEvent(QX11Info::display(), QX11Info::appRootWindow(), false,
               SubstructureNotifyMask | SubstructureRedirectMask,
               (XEvent*) &e);
}
#endif

void App::forceActivate(QWidget *window)
{
    if (!window)
        window = topWindow();

    if (window)
    {
        qDebug() << "calling activate()";
        window->raise();
        window->activateWindow();
#ifdef Q_WS_X11
        activateWindowX11(window);
#endif
    }
}

namespace {

class WsWindowFactory : public moo::DocumentManager::WindowFactory {
    moo::DocumentWindow *createWindow() const
    {
        return new WsWindow(false);
    }
};

}

App::App(int &argc, char **argv) :
    moo::App(argc, argv,
             "ggap.sourceforge.net",
             "ggap", GGAP_VERSION)
{
#if !defined(Q_OS_MAC)
    QIcon icon;
    icon.addFile(":/icons/ggap/themed/16x16/apps/ggap.png", QSize(16, 16));
    icon.addFile(":/icons/ggap/themed/32x32/apps/ggap.png", QSize(32, 32));
    icon.addFile(":/icons/ggap/themed/48x48/apps/ggap.png", QSize(48, 48));
    icon.addFile(":/icons/ggap/themed/64x64/apps/ggap.png", QSize(64, 64));
    setWindowIcon(icon);
#endif

    QDesktopServices::setUrlHandler("ghelp", this, "gapHelp");
    QDesktopServices::setUrlHandler("gfile", this, "gapHelp");
    QDesktopServices::setUrlHandler("help", this, "gapHelp");

    moo::DocumentManager::setWindowFactory(new WsWindowFactory);
}


static void deleteFile(const QString &name)
{
    QFileInfo fi(name);
    if (fi.isDir() && !fi.isSymLink())
    {
        QDir dir(name);
        QStringList files = dir.entryList();
        foreach (const QString &f, files)
            if (f != "." && f != "..")
                deleteFile(dir.filePath(f));
        QDir(fi.path()).rmdir(fi.fileName());
    }
    else
        QFile::remove(name);
}

App::~App()
{
}

void App::aboutDialog()
{
    AboutDialog::showDialog();
}

void App::prefsDialog()
{
    PrefsDialog::showDialog(0);
}

void App::gapHelp(const QUrl &url)
{
    HelpBrowser::showWindow(url);
}

void App::checkGap()
{
    static bool been_here;

    if (been_here)
        return;

    been_here = true;

    if (GapOptions().exe().isEmpty())
    {
        moo::dlg::information(0, QString(), tr("GAP not found"),
                              tr("Please select the root GAP directory or GAP "
                                 "executable in the Preferences dialog"));
        PrefsDialog::execDialog(0);
    }
}

int App::exec()
{
    checkGap();

    foreach (const QString &script, appOptions.scripts)
        runScript(script);

    foreach (const QString &f, appOptions.files_to_open)
        loadFile(f);

    {
        // XXX
        bool hasDocumentWindow = false;
        QWidgetList windows = topLevelWidgets();

        foreach (QWidget *w, windows)
        {
            if ((w = dynamic_cast<WsWindow*>(w)))
            {
                hasDocumentWindow = true;
                break;
            }
        }

        if (!hasDocumentWindow)
        {
            WsWindow *w = new WsWindow(false);
            w->present(0);
        }
    }

    if (appOptions.raise)
    {
        QWidget *win = 0;

        if (!(win = activeWindow()))
        {
            QWidgetList windows = topLevelWidgets();
            foreach (QWidget *w, windows)
                if ((win = dynamic_cast<WsWindow*>(w)))
                    break;
        }

        if (win)
        {
            win->raise();
            win->activateWindow();
        }
    }

    if (!appOptions.help_url.isEmpty())
        HelpBrowser::showWindow(appOptions.help_url);
    else if (appOptions.open_help_browser)
        HelpBrowser::showWindow();

    return QApplication::exec();
}

void App::loadFile(const QString &filename)
{
    WsWindow::openFile(filename);
}

void App::messageReceived(const QByteArray &message)
{
    qDebug() << "messageReceived: " << message;
    if (message.startsWith("script:"))
        runScript(QString::fromUtf8(message.data() + strlen("script:"),
                                    message.length() - strlen("script:")));
    else if (message.startsWith("appmsg:"))
        doAppMsg(message.mid(strlen("appmsg:")));
    else
        qCritical("unknown message type: %s", message.data());
}

void App::runScript(const QString &script)
{
    ScriptRunner sr;
    sr.runScript(script);
}

// void App::runScriptFile(const QString &filename)
// {
// }
